@@ -27,7 +27,7 @@
 #include <config.h>
 #endif
 
-#include <libgnome/gnome-i18n.h>
+#include <glib/gi18n.h>
 #include <string.h>
 #include <gnome.h>
 #include <bonobo.h>
@@ -70,7 +70,7 @@
 #include "text.h"
 #include "paragraph.h"
 #include "body.h"
-#include "spell.h"
+#include "spellchecker.h"
 #include "html-stream-mem.h"
 
 #include "gtkhtmldebug.h"
@@ -95,18 +95,20 @@ typedef struct _SetFrameData SetFrameData;
 static GtkHTMLEditorAPI *editor_api;
 
 static void
-activate_cb (BonoboControl      *control,
+activate_ui_cb (BonoboControl      *control,
 	     gboolean            active,
 	     GtkHTMLControlData *cd)
 {
-	Bonobo_UIContainer remote_ui_container;
 	BonoboUIComponent *ui_component;
 
-	printf ("ACTIVATE\n");
+	ui_component = bonobo_control_get_ui_component (control);
 
 	if (active) {
+		Bonobo_UIContainer remote_ui_container;
+
+		cd->uic = ui_component;
+
 		remote_ui_container = bonobo_control_get_remote_ui_container (control, NULL);
-		cd->uic = ui_component = bonobo_control_get_ui_component (control);
 		bonobo_ui_component_set_container (ui_component, remote_ui_container, NULL);
 		bonobo_object_release_unref (remote_ui_container, NULL);
 
@@ -118,13 +120,13 @@ static void
 set_frame_cb (BonoboControl *control,
 	      gpointer data)
 {
-	Bonobo_UIContainer remote_ui_container;
-	BonoboUIComponent *ui_component;
 	GtkHTMLControlData *control_data;
 	GtkWidget *toolbar;
 	GtkWidget *scrolled_window;
 	Bonobo_ControlFrame frame;
 
+	/* printf ("set_frame_cb\n"); */
+
 	control_data = (GtkHTMLControlData *) data;
 
 	frame = bonobo_control_get_control_frame (control, NULL);
@@ -133,9 +135,7 @@ set_frame_cb (BonoboControl *control,
 
 	CORBA_Object_release (frame, NULL);
 
-	remote_ui_container = bonobo_control_get_remote_ui_container (control, NULL);
-	control_data->uic = ui_component = bonobo_control_get_ui_component (control);
-	bonobo_ui_component_set_container (ui_component, remote_ui_container, NULL);
+	control_data->uic = bonobo_control_get_ui_component (control);
 
 	/* Setup the tool bar.  */
 
@@ -150,19 +150,10 @@ set_frame_cb (BonoboControl *control,
 
 	gtk_box_pack_start (GTK_BOX (control_data->vbox), scrolled_window, TRUE, TRUE, 0);
 
-	/* Setup the menu bar.  */
-
-	menubar_setup (ui_component, control_data);
-
-	if (!spell_has_control ()) {
-		control_data->has_spell_control = FALSE;
-		bonobo_ui_component_set_prop (ui_component, "/commands/EditSpellCheck", "sensitive", "0", NULL);
-	} else
-		control_data->has_spell_control = TRUE;
+	/* hack to setup the bonobo ui if activate isn't working */
+	activate_ui_cb (control, TRUE, control_data);
 
 	gtk_html_set_editor_api (GTK_HTML (control_data->html), editor_api, control_data);
-
-	bonobo_object_release_unref (remote_ui_container, NULL);
 }
 
 static gint
@@ -184,8 +175,8 @@ release (GtkWidget *widget, GdkEventButton *event, GtkHTMLControlData *cd)
 			;
 		}
 		if (run_dialog) {
-			cd->properties_dialog = gtk_html_edit_properties_dialog_new (cd, FALSE, _("Properties"), 
-										     ICONDIR "/properties-16.png");
+			cd->properties_dialog = gtk_html_edit_properties_dialog_new (cd, _("Properties"), 
+										     gnome_icon_theme_lookup_icon (cd->icon_theme, "stock_properties", 16, NULL, NULL));
 			html_cursor_jump_to (e->cursor, e, cd->obj, 0);
 			html_engine_disable_selection (e);
 			html_engine_set_mark (e);
@@ -197,7 +188,6 @@ release (GtkWidget *widget, GdkEventButton *event, GtkHTMLControlData *cd)
 				gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
 									   GTK_HTML_EDIT_PROPERTY_IMAGE, _("Image"),
 									   image_properties,
-									   image_apply_cb,
 									   image_close_cb);
 				start = GTK_HTML_EDIT_PROPERTY_IMAGE;
 				break;
@@ -206,7 +196,6 @@ release (GtkWidget *widget, GdkEventButton *event, GtkHTMLControlData *cd)
 				gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
 									   GTK_HTML_EDIT_PROPERTY_TEXT, _("Text"),
 									   text_properties,
-									   text_apply_cb,
 									   text_close_cb);
 				start = (HTML_OBJECT_TYPE (cd->obj) == HTML_TYPE_TEXT)
 					? GTK_HTML_EDIT_PROPERTY_TEXT
@@ -217,7 +206,6 @@ release (GtkWidget *widget, GdkEventButton *event, GtkHTMLControlData *cd)
 				gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
 									   GTK_HTML_EDIT_PROPERTY_RULE, _("Rule"),
 									   rule_properties,
-									   rule_apply_cb,
 									   rule_close_cb);
 				start = GTK_HTML_EDIT_PROPERTY_RULE;
 				break;
@@ -227,12 +215,10 @@ release (GtkWidget *widget, GdkEventButton *event, GtkHTMLControlData *cd)
 			gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
 								   GTK_HTML_EDIT_PROPERTY_PARAGRAPH, _("Paragraph"),
 								   paragraph_properties,
-								   paragraph_apply_cb,
 								   paragraph_close_cb);
 			gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
 								   GTK_HTML_EDIT_PROPERTY_BODY, _("Page"),
 								   body_properties,
-								   body_apply_cb,
 								   body_close_cb);
 			gtk_html_edit_properties_dialog_show (cd->properties_dialog);
 			gtk_html_edit_properties_dialog_set_page (cd->properties_dialog, start);
@@ -251,12 +237,17 @@ load_from_file (GtkHTML *html,
 	unsigned char buffer[4096];
 	int len;
 	int fd;
-        const char *path;
+        char *path;
 
         if (strncmp (url, "file:", 5) == 0)
-		path = url + 5; 
+		path = g_filename_from_uri(url, NULL, NULL);
+	else
+		path = g_strdup(url);
 
-	if ((fd = open (path, O_RDONLY)) == -1) {
+	fd = open (path, O_RDONLY);
+	g_free(path);
+
+	if (fd == -1) {
 		g_warning ("%s", g_strerror (errno));
 		return FALSE;
 	}
@@ -367,10 +358,8 @@ html_button_pressed_after (GtkWidget *html, GdkEventButton *event, GtkHTMLContro
 			html_engine_select_interval (e, html_interval_new_from_cursor (e->mark, e->cursor));
 			/* printf ("delete template text\n"); */
 			html_engine_delete (cd->html->engine);
-		} else if (HTML_IS_IMAGE (obj) && html_object_get_data (obj->parent, "template_image")) {
+		} else if (HTML_IS_IMAGE (obj) && html_object_get_data (obj->parent, "template_image"))
 			property_dialog_show (cd);
-			html_object_set_data_full (obj->parent, "template_image", NULL, NULL);
-		}
 	}
 
 	return FALSE;
@@ -392,9 +381,6 @@ editor_init_painters (GtkHTMLControlData *cd)
 		cd->gdk_painter = HTML_GDK_PAINTER (html->engine->painter);
 		cd->plain_painter = HTML_GDK_PAINTER (html_plain_painter_new (GTK_WIDGET (html), TRUE));
 
-		html_colorset_add_slave (html->engine->settings->color_set, 
-					 HTML_PAINTER (cd->plain_painter)->color_set);
-
 		/* the plain painter starts with a ref */
 		g_object_ref (G_OBJECT (cd->gdk_painter));
 	}	
@@ -445,7 +431,7 @@ editor_set_format (GtkHTMLControlData *cd, gboolean format_html)
 		
 }
 
-static enum {
+typedef enum {
 	PROP_EDIT_HTML,
 	PROP_HTML_TITLE,
 	PROP_INLINE_SPELLING,
@@ -456,7 +442,7 @@ static enum {
 static void
 editor_get_prop (BonoboPropertyBag *bag,
 		 BonoboArg         *arg,
-		 guint              arg_id,
+		 EditorControlProps arg_id,
 		 CORBA_Environment *ev,
 		 gpointer           user_data)
 {
@@ -534,7 +520,6 @@ editor_control_construct (BonoboControl *control, GtkWidget *vbox)
 	html_widget = gtk_html_new ();
 	gtk_html_load_empty (GTK_HTML (html_widget));
 	gtk_html_set_editable (GTK_HTML (html_widget), TRUE);
-	gtk_html_set_animate (GTK_HTML (html_widget), FALSE);
 
 	cd = gtk_html_control_data_new (GTK_HTML (html_widget), vbox);
 	g_signal_connect (control, "destroy", G_CALLBACK (control_destroy), cd);
@@ -632,7 +617,7 @@ editor_control_construct (BonoboControl *control, GtkWidget *vbox)
 	   embedded in its control frame.  We use the "set_frame" signal to
 	   handle that.  */
 
-	g_signal_connect (control, "activate", G_CALLBACK (activate_cb), cd);
+	//g_signal_connect (control, "activate", G_CALLBACK (activate_ui_cb), cd);
 	g_signal_connect (control, "set_frame", G_CALLBACK (set_frame_cb), cd);
 	g_signal_connect (html_widget, "url_requested", G_CALLBACK (url_requested_cb), cd);
 	g_signal_connect (html_widget, "button_press_event", G_CALLBACK (html_button_pressed), cd);
@@ -670,10 +655,15 @@ send_event_str (GNOME_GtkHTML_Editor_Engine engine, GNOME_GtkHTML_Editor_Listene
 {
 	CORBA_Environment ev;
 	GValue *gvalue_retval = NULL;
-	BonoboArg *bonobo_arg = bonobo_arg_new (bonobo_arg_type_from_gtype (G_VALUE_TYPE (arg)));
+	BonoboArg *bonobo_arg;
 	BonoboArg *bonobo_retval;
 
-	bonobo_arg_from_gvalue (bonobo_arg, arg);
+	bonobo_arg = bonobo_arg_new (bonobo_arg_type_from_gtype (G_VALUE_TYPE (arg)));
+
+	/* bonobo_arg_from_gvalue() doesn't handle the case of a NULL string
+	   properly, at least in libbonobo 2.4.0.  */
+	if (! G_VALUE_HOLDS_STRING (arg) || g_value_get_string (arg) != NULL)
+		bonobo_arg_from_gvalue (bonobo_arg, arg);
 
 	/* printf ("sending to listener\n"); */
 	CORBA_exception_init (&ev);
