@@ -22,14 +22,34 @@
 #ifndef _HTMLENGINE_EDIT_RULE_H
 #define _HTMLENGINE_EDIT_RULE_H
 
-#include "htmlengine.h"
-#include "htmlobject.h"
+#include "htmltypes.h"
+#include "htmlenums.h"
 
-void html_engine_insert_rule (HTMLEngine      *e,
-			      gint            length,
-			      gint            percent,
-			      gint            size,
-			      gboolean        shade,
-			      HTMLHAlignType  halign);
+void  html_engine_insert_rule  (HTMLEngine     *e,
+				gint            length,
+				gint            percent,
+				gint            size,
+				gboolean        shade,
+				HTMLHAlignType  halign);
+void  html_rule_set            (HTMLRule       *rule,
+				HTMLEngine     *e,
+				gint            length,
+				gint            percent,
+				gint            size,
+				gboolean        shade,
+				HTMLHAlignType  halign);
+void  html_rule_set_length     (HTMLRule       *rule,
+				HTMLEngine     *e,
+				gint            length,
+				gint            percent);
+void  html_rule_set_size       (HTMLRule       *rule,
+				HTMLEngine     *e,
+				gint            size);
+void  html_rule_set_shade      (HTMLRule       *rule,
+				HTMLEngine     *e,
+				gboolean        shade);
+void  html_rule_set_align      (HTMLRule       *rule,
+				HTMLEngine     *e,
+				HTMLHAlignType  halign);
 
 #endif
